Require Import Coq.Program.Equality.
Require Import Coq.Program.Wf.
Require Import LibTactics.
Require Import InfiniteTypes.
Require Import Wf.
Require Import ArithMisc.
Set Implicit Arguments.

(* This ordering is used in the proof that the model validates the subtyping
   relation: see [subtyping_valid_preliminary] in [Model.v]. It is a
   well-founded ordering, but the argument is somewhat indirect. *)

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Everything is parameterized over a natural integer [q], which plays the role
   of an upper bound for the index [j] -- see the definitions below. *)

Section Q.

Variable q : nat.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We are interested in relations over quadruples of two natural integers and
   two types. *)

Definition quadruple :=
  (nat * ty * nat * ty)%type.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The relation we are ultimately interested in, [svo], is the union of four
   sub-relations. *)

(* Bullet left. *)

Inductive BL: quadruple -> quadruple -> Prop :=
| BulletLeft:
    forall i t j u,
    wf (TyBullet t) -> wf u ->
    finite (TyBullet t) -> finite u ->
    j < q ->
    BL (i + 1, t, j, u) (i, TyBullet t, j, u).

(* Bullet right. *)

Inductive BR: quadruple -> quadruple -> Prop :=
| BulletRight:
    forall i t j u,
    wf t -> wf (TyBullet u) ->
    finite t -> finite (TyBullet u) ->
    j < q ->
    BR (i, t, j + 1, u) (i, t, j, TyBullet u).

(* Arrow left. *)

Inductive AL: quadruple -> quadruple -> Prop :=
| ArrowLeft:
    forall i k t1 t2 j u1 u2,
    wf (TyArrow t1 t2) -> wf (TyArrow u1 u2) ->
    finite (TyArrow t1 t2) -> finite (TyArrow u1 u2) ->
    j < q ->
    AL (j, u1, i + k, t1) (i, TyArrow t1 t2, j, TyArrow u1 u2).

(* Arrow right. *)

Inductive AR: quadruple -> quadruple -> Prop :=
| ArrowRight:
    forall i k t1 t2 j u1 u2,
    wf (TyArrow t1 t2) -> wf (TyArrow u1 u2) ->
    finite (TyArrow t1 t2) -> finite (TyArrow u1 u2) ->
    j < q ->
    AR (i + k, t2, j, u2) (i, TyArrow t1 t2, j, TyArrow u1 u2).

Hint Constructors BL BR AL AR.

(* The union. *)

Inductive svo : quadruple -> quadruple -> Prop :=
| SvoBL: forall a b, BL a b -> svo a b
| SvoBR: forall a b, BR a b -> svo a b
| SvoAL: forall a b, AL a b -> svo a b
| SvoAR: forall a b, AR a b -> svo a b.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* It is impossible to go forever using only AL and AR. This is due to the
   fact that types are well-formed: every infinite path must eventually cross
   a bullet. *)

Lemma not_always_arrows_preliminary:
  forall t,
  bb t ->
  forall s i j u,
  ~ always (union (rel AL) (rel AR)) (SeqCons s (i, t, j, u)) /\
  ~ always (union (rel AL) (rel AR)) (SeqCons s (i, u, j, t)).
Proof.
  induction 1; intros;
  destruct s as [ s [[[ i1 t1 ] j1 ] u1 ]];
  split;
  intro h;
  generalize (A_cons_head h); intros [ head | head ]; unfold rel in head; dependent destruction head;
  generalize (A_cons_tail h).
  (* Arrow. *)
    (* Left goal. *)
      (* Left. *)
      eapply IHbb1.
      (* Right. *)
      eapply IHbb2.
    (* Right goal. *)
      (* Left. *)
      eapply IHbb1.
      (* Right. *)
      eapply IHbb2.
Qed.

Lemma not_always_arrows:
  forall s,
  ~ always (union (rel AL) (rel AR)) s.
Proof.
  intros s.
  dependent destruction s. destruct q0 as [[[ i t ] j ] u ].
  intro h.
  generalize (A_cons_head h); intro head.
  generalize h. clear h.
  eapply not_always_arrows_preliminary.
  (* There remains to show that [u] is bullet-bounded. *)
  destruct head as [ head | head ];
  destruct s as [ s a ];
  unfold rel in head;
  dependent destruction head;
  eauto.
Qed.

Lemma often_bullets:
  subset
    (always (rel svo))
    (always (eventually (union (rel BL) (rel BR)))).
Proof.
  apply always_often.
  assert (
    subset
      (intersection (rel svo) (neg (union (rel BL) (rel BR))))
      (union (rel AL) (rel AR))
  ) as fact.
    unfold subset, intersection, rel, neg, union.
    intros [[ ? a ] b ]. intros [ hsvo h ]. dependent destruction hsvo; intuition eauto.
  rewrite fact.
  unfold subset, empty. eapply not_always_arrows.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A property of a quadruple: ``the sum of [i] and [j] is at least [n]''. *)

Definition over n (a : quadruple) :=
  match a with (i, t, j, u) => i + j >= n end.

(* This is lifted to a property of (the head of) a sequence. *)

Definition now_over k :=
  now (over k).

(* The larger [n], the more precise the property. *)

Lemma now_over_monotonic:
  forall m n,
  m <= n ->
  subset (now_over n) (now_over m).
Proof.
  unfold now_over, now, over.
  intros ? ? ? [ s [[[ i t ] j ] u ]] ?.
  omega.
Qed.

(* The property is preserved by every step, and strictly refined by [BL] and
   [BR] steps. *)

Lemma over_preserved:
  forall a1 a2 n,
  over n a2 ->
  svo a1 a2 ->
  over n a1.
Proof.
  unfold over.
  intros [[[ i1 t1 ] j1 ] u1 ]  [[[ i2 t2 ] j2 ] u2 ] ? ? h.
  dependent destruction h; dependent destruction H0; omega.
Qed.

Lemma over_increased:
  forall a1 a2 n,
  over n a2 ->
  BL a1 a2 \/ BR a1 a2 ->
  over (n + 1) a1.
Proof.
  unfold over.
  introv ? aa.
  destruct aa as [ h | h ]; dependent destruction h; omega.
Qed.

Lemma over_increased_pedantic:
  forall n,
  ss
    (intersection (now (over n)) (union (rel BL) (rel BR)))
    (later (now (over (n + 1)))).
Proof.
  unfold intersection, now, union, rel, later.
  intros ? [ [ s a ] b ] [ ? ? ].
  eapply over_increased; eauto.
Qed.

(* If the property holds now, then it holds forever. *)

Lemma now_over_preservation:
  forall n s,
  now_over n s ->
  always (rel svo) s ->
  always (now_over n) s.
Proof.
  cofix.
  intros ? [[ s b ] a ] ? h.
  constructor.
  apply now_over_preservation. apply (@over_preserved b a). auto. forwards: A_cons_head. eassumption. auto.
  eapply A_cons_tail. eassumption.
  auto.
Qed.

(* If the property holds now, then a strictly more precise property
   will hold at some point in the future. This exploits the fact that
   there must a [BL] or [BR] step every so often. *)

Lemma now_over_step:
  forall n s,
  now_over n s ->
  always (rel svo) s ->
  eventually (now_over (n + 1)) s.
Proof.
  introv ? asvo.
  (* At some point, we must use BL or BR. *)
  forwards bb: (A_intro (often_bullets asvo)).
  (* We are over [n] forever. *)
  forwards on: now_over_preservation. eassumption. eassumption.
  (* Combine these facts. *)
  forwards fact: A_and_E. eexact on. eexact bb.
  (* There! *)
  eapply to_ss. eexact fact.
  rewrite (@over_increased_pedantic n).
  rewrite EL.
  auto.
Qed.

(* As a result of the above properties, the sum [i + j] must tend towards
   infinity. *)

Lemma now_over_increasingly:
  subset
    (always (rel svo))
    (increasingly now_over).
Proof.
  assert (
    subset (always (rel svo)) (intersection (always (rel svo)) (now_over 0))
  ) as obvious.
    unfold subset, intersection, now_over, now, over.
    intros [ s [[[ i t ] j ] u ]] ?.
    intuition eauto. omega.
  rewrite obvious.
  apply step_increasingly.
  apply now_over_monotonic.
  apply now_over_preservation.
  apply now_over_step.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now show that, because the sum [i + j] tends to infinity, after a while,
   there cannot be any [AL] step. Indeed, such a step reverses [i] and [j], but
   if [i] has grown too large, this will make [j] too large and prevent the
   next step. *)

Lemma j_bounded:
  forall a i j t u,
  svo a (i, t, j, u) ->
  j < q.
Proof.
  introv h;
  dependent destruction h;
  dependent destruction H;
  auto.
Qed.

Lemma no_AL_preliminary:
  subset
    (intersection (always (now_over (2 * q))) (intersection (always (rel svo)) (rel AL)))
    (@empty quadruple).
Proof.
  intros [[[ s c ] b ] a ] [ h2 [ h1 h3 ]]. unfold empty.
  (* We consider an AL step from a to b, followed with an arbitrary svo step
     from b to c. *)
  assert (rel svo (SeqCons (SeqCons s c) b)) as h.
    eapply A_cons_head.
    eapply A_cons_tail.
    eassumption.
  unfold rel in *.
  dependent destruction h3.        (* exploit the AL step; introduces j < q *)
  generalize (j_bounded h); intro. (* exploit the next step; introduces i + k < q, hence i < q *)
  dependent destruction h2.        (* exploit the hypothesis i + j >= 2q *)
  unfold now_over, now, over in *.
  omega.                           (* contradiction *)
Qed.

Lemma no_AL_preliminary_2:
  subset
    (intersection (intersection (always (rel svo)) (eventually (rel AL))) (always (now_over (2 * q)))) (@empty quadruple).
Proof.
  rewrite IAE.
  rewrite I_commutative.
  rewrite IAE.
  rewrite <- E_false.
  apply E_covariant.
  apply no_AL_preliminary.
Qed.

Lemma no_AL:
  subset
    (always (rel svo))
    (eventually (always (neg (rel AL)))).
Proof.
  rewrite <- NE.
  rewrite <- NA.
  apply subset_neg.
  intros s [ asvo aeAL ].
  (* Eventually, the sum i + j will be over 2q. *)
  generalize (now_over_increasingly asvo (2 * q)); intro eo2q.
  (* Sometime after this point, an AL step must occur. Combine
     what is known there. *)
  generalize (A_and_E (A_and_A (AA asvo) aeAL) eo2q); intro step.
  (* Go to this point. *)
  eapply E_false_variant. eexact step.
  apply no_AL_preliminary_2.
Qed.

Lemma svo_arblbr:
  subset
    (always (rel svo))
    (eventually (always (union (rel AR) (union (rel BL) (rel BR))))).
Proof.
  eapply (AEA_reinforce_variant no_AL).
  unfold union, rel, neg.
  intros [[ s b ] a ] [ hsvo ? ].
  dependent destruction hsvo; intuition eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now know that, after a while, there cannot be any [AL] step. As a
   result, the indices [i] and [j] are no longer exchanged. This implies, in
   particular, that [j] grows with time, and grows strictly at every [BR]
   step. Because [j] is bounded, after a while, there cannot be any [BR]
   step. *)

(* A property of a quadruple: ``[j] is at least [n], and, if [n] is larger
   than [q], then contradiction''. *)

Definition jover n (a : quadruple) :=
  match a with (i, t, j, u) => j >= n /\ n <= q end.

(* This is lifted to a property of (the head of) a sequence, and combined
   with the assumption that [BR] occurs infinitely often. *)

Definition now_jover k :=
  intersection (now (jover k)) (always (eventually (rel BR))).

Hint Unfold now_jover.

(* The larger [n], the more precise the property. *)

Lemma now_jover_monotonic:
  forall m n,
  m <= n ->
  subset (now_jover n) (now_jover m).
Proof.
  unfold now_jover, now, jover, intersection.
  intros ? ? ? [ s [[[ i t ] j ] u ]] ?.
  intuition omega.
Qed.

(* The property is preserved by every step, and strictly refined by [BR]
   steps. *)

Lemma jover_preserved:
  forall a1 a2 n,
  jover n a2 ->
  AR a1 a2 \/ BL a1 a2 \/ BR a1 a2 ->
  jover n a1.
Proof.
  unfold jover.
  introv ? aa.
  destruct aa as [ h | [ h | h ]]; dependent destruction h; omega.
Qed.

Lemma jover_increased:
  forall a1 a2 n,
  jover n a2 ->
  BR a1 a2 ->
  jover (n + 1) a1.
Proof.
  unfold jover.
  introv ? h.
  dependent destruction h.
  omega.
Qed.

Lemma jover_increased_pedantic:
  forall n,
  ss
    (intersection (now_jover n) (rel BR))
    (later (now_jover (n + 1))).
Proof.
  unfold now_jover, intersection, now, union, rel, later.
  intros ? [ [ s a ] b ] [ [ ? ? ] ? ].
  split.
  eapply jover_increased; eauto.
  eapply A_cons_tail; eauto.
Qed.

(* If the property holds now, then it holds forever. *)

Lemma now_jover_preservation:
  forall n s,
  now_jover n s ->
  always (union (rel AR) (union (rel BL) (rel BR))) s ->
  always (now_jover n) s.
Proof.
  cofix.
  intros ? [[ s b ] a ] [ ? ? ] h.
  constructor.
  apply now_jover_preservation. 
  split.
    apply (@jover_preserved b a). auto. forwards: A_cons_head. eassumption. auto.
    eapply A_cons_tail. eassumption.
    eapply A_cons_tail. eassumption.
  auto.
Qed.

(* If the property holds now, then a strictly more precise property
   will hold at some point in the future. This exploits the fact that
   there must a [BR] step every so often. *)

Lemma now_jover_step:
  forall n s,
  now_jover n s ->
  always (union (rel AR) (union (rel BL) (rel BR))) s ->
  eventually (now_jover (n + 1)) s.
Proof.
  introv [ ? aeBR ] asvo.
  (* At some point, we must use BR. *)
  (* We are over [n] forever. *)
  forwards on: now_jover_preservation. eauto. eassumption.
  (* Combine these facts. *)
  forwards fact: A_and_E. eexact on. eexact (A_intro aeBR).
  (* There! *)
  eapply to_ss. eexact fact.
  rewrite (@jover_increased_pedantic n).
  rewrite EL.
  auto.
Qed.

(* As a result of the above properties, if [BR] occurs infinitely often, then
   [j] must tend towards infinity. *)

Lemma now_jover_increasingly:
  subset
    (intersection
      (always (union (rel AR) (union (rel BL) (rel BR))))
      (always (eventually (rel BR))))
    (increasingly now_jover).
Proof.
  assert (
    subset
      (always (eventually (rel BR)))
      (now_jover 0)
  ) as obvious.
    unfold subset, now_jover, intersection, now, jover.
    intros [ s [[[ i t ] j ] u ]] ?.
    intuition omega.
  rewrite obvious.
  apply step_increasingly.
  apply now_jover_monotonic.
  apply now_jover_preservation.
  apply now_jover_step.
Qed.

(* But this is impossible, so [BR] cannot occur infinitely often. *)

Lemma no_BR:
  subset
    (always (union (rel AR) (union (rel BL) (rel BR))))
    (eventually (always (neg (rel BR)))).
Proof.
  rewrite <- NE.
  rewrite <- NA.
  apply subset_neg.
  intros s asvo_aeBR.
  (* Eventually, [j] will be over [q]. *)
  generalize (now_jover_increasingly asvo_aeBR (q + 1)); intro jo.
  (* Go there. *)
  eapply E_false_variant. eexact jo.
  rewrite <- A_false.
  apply A_covariant.
  (* This boils down to proving that [now_jover (q + 1)] is empty. *)
  unfold now_jover.
  intros [ ? [[[ i t ] j ] u ]] [ h _ ].
  unfold now, jover in h.
  unfold empty. omega.
Qed.

Lemma arblbr_arbl:
  subset
    (always (union (rel AR) (union (rel BL) (rel BR))))
    (eventually (always (union (rel AR) (rel BL)))).
Proof.
  eapply (AEA_reinforce_variant no_BR).
  unfold union, rel, neg.
  intros [[ s b ] a ] [ ? ? ].
  intuition eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* However, it is impossible to go forever using only AR and BL. This is due
  to the finiteness of [t] and the well-formedness of [u]. *)

Lemma not_always_arbl_preliminary:
  forall t,
  tf t ->
  forall u,
  bb u ->
  forall s i j,
  ~ always (union (rel AR) (rel BL)) (SeqCons s (i, t, j, u)).
Proof.
  induction 1; induction 1; intros; destruct s as [ s a ];
  intro h; generalize (A_cons h); intros [[ head | head ] tail ]; unfold rel in head; dependent destruction head;
  unfold not in *; eauto.
Qed.

Lemma not_always_arbl:
  forall s,
  ~ always (union (rel AR) (rel BL)) s.
Proof.
  intros [ s [[[ i t ] j ] u ] ].
  intro h.
  generalize (A_cons_head h); intro head.
  generalize h. clear h.
  apply not_always_arbl_preliminary;
  (* There remains to show that [t] is finite and [u] is well-formed. *)
  destruct head as [ head | head ];
  destruct s as [ s a ];
  unfold rel in head;
  dependent destruction head;
  eauto.
Qed.

Lemma always_arbl_empty:
  subset
    (always (union (rel AR) (rel BL)))
    (@empty quadruple).
Proof.
  intros s ?. eapply not_always_arbl. eassumption.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now conclude. *)

Lemma always_svo_empty:
  subset
    (always (rel svo))
    (@empty quadruple).
Proof.
  rewrite svo_arblbr.
  rewrite arblbr_arbl.
  rewrite always_arbl_empty.
  rewrite E_false.
  rewrite E_false.
  auto.
Qed.

Theorem svo_wf:
  well_founded svo.
Proof.
  apply NNPP. intro h.
  generalize (if_sequence h); intros [ s asvo ].
  eapply always_svo_empty. eassumption.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

End Q.

